/*
 * Author:
 * Anthony Stahler
 * unfair@unfairnet.com
 *
 * Please report any bugs to http://bugs.unfairnet.com
 */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02111-1301  USA
 */

#define PURPLE_PLUGINS

#include <glib.h>

#ifndef G_GNUC_NULL_TERMINATED
#  if __GNUC__ >= 4
#    define G_GNUC_NULL_TERMINATED __attribute__((__sentinel__))
#  else
#    define G_GNUC_NULL_TERMINATED
#  endif /* __GNUC__ >= 4 */
#endif /* G_GNUC_NULL_TERMINATED */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <stdlib.h>

#include <errno.h> // socket error types

#ifndef _WIN32
#	include <sys/socket.h>
#else
#	include "libc_interface.h"
// #include <winsock.h>
#	define EWOULDBLOCK WSAEWOULDBLOCK
#endif

#include <unistd.h> // close socket
#include "cipher.h" // hash password

#include "prpl.h"
#include "notify.h"
#include "plugin.h"
#include "version.h"

#include "accountopt.h"
#include "connection.h"
#include "debug.h"
#include "dnsquery.h"
#include "proxy.h"
#include "request.h"
#include "account.h"
#include "blist.h"
#include "cmds.h"
#include "conversation.h"
#include "privacy.h"
#include "roomlist.h"
#include "status.h"
#include "util.h"


// RSA STUFF
#include "rsa.h"
#include "x509.h"

typedef struct	_UNetData		UNetData;
typedef struct	_ustr			ustr;
typedef struct	_ApproveWrapper	ApproveWrapper;

static void init_plugin(PurplePlugin *plugin);
static void destroy_plugin(PurplePlugin *plugin);
static gboolean plugin_load(PurplePlugin *plugin);
static gboolean plugin_unload(PurplePlugin *plugin);
static gchar *buddy_status(PurpleBuddy *buddy);
static GList *buddy_status_types(PurpleAccount *account);
static GList *buddy_rclick_menu(PurpleBlistNode *node);
static void user_login(PurpleAccount *acct);
static void user_logout(PurpleConnection *pc);
static int buddy_im(PurpleConnection *pc, const char *who, const char *message, PurpleMessageFlags flags);
static void buddy_get_info(PurpleConnection *pc, const char *who);
static void buddy_set_status(PurpleAccount *account, PurpleStatus *status);
static void buddy_add(PurpleConnection *pc, PurpleBuddy *buddy, PurpleGroup *group);
static void buddy_remove(PurpleConnection *pc, PurpleBuddy *buddy, PurpleGroup *group);
static void buddy_block(PurpleConnection *pc, const char *name);
static void buddy_unblock(PurpleConnection *pc, const char *name);
static void buddy_join_chat(PurpleConnection *pc, GHashTable *components);
static void buddy_leave_chat(PurpleConnection *pc, int id);
static int buddy_chat_im(PurpleConnection *pc, int id, const char *message, PurpleMessageFlags flags);
static gboolean buddy_offline_support(const PurpleBuddy *buddy);
static const char *base_icon(PurpleAccount *account, PurpleBuddy *buddy);
static void update_connection_step(const char *conn_text, PurpleAccount *acct);
static void add_buddy(const char *buddy_name, const char *group_name, PurpleAccount *acct);
static void remove_buddy(const char *buddy_name, const char *group_name, PurpleAccount *acct);
static UNetData *fetch_node_ud(PurpleBlistNode *node);
static int send_im(const char *touser, const char *message, PurpleAccount *acct);
static void login(const char *user, const char *pass, PurpleAccount *acct);
static void request_buddies(PurpleAccount *acct);
static void uconnect(PurpleAccount *acct, const char *server, int port);
static void uconnect_cb(gpointer userdata, gint fd, const gchar *error_message);
static void incoming_data(gpointer data, gint source, PurpleInputCondition cond);
static gboolean handle_packet(PurpleAccount *acct, unsigned char *packet, int plen);
static int two_bytes_to_int(const unsigned char *buffer);
static int one_byte_to_int(unsigned char tmpchr);
static ustr *construct_packet(PurpleAccount *acct, int encrypted, int ptype, int stype, int argc, ...);
static ustr *construct_subpacket(int sptype, const char *data, int datalen, int lenbyte);
static gboolean request_salt(gpointer data);
static void int_two_bytes(int num, char *rtnval, int pos);
static char *unet_hash(PurpleAccount *acct, char *ciphername, int hashbytes);
static int fetch_segment(const unsigned char *data, int plen, unsigned char **target, int offset, gboolean nullterm, int lenbyte);
static gboolean send_packet(PurpleAccount *acct, ustr *data);
static void close_socket(PurpleAccount *acct);
static gboolean send_keepalive(gpointer data);
static gboolean send_packet_finish(gpointer data);
static char *enhex(guchar *tmpstr, int length);
static void update_buddy_status(int status_type, char *buddyname, char *status_message, time_t mtime, PurpleAccount *acct);
static void buddy_tooltip(PurpleBuddy *pb, PurpleNotifyUserInfo *user_info, gboolean full);
static void clear_buddy_list(gpointer data, gpointer userdata);
static void close_socket_destroy(UNetData *ud);
static int send_chat(const char *tochat, const char *message, PurpleAccount *acct);
static GList *unet_chat_info(PurpleConnection *gc);
static GHashTable *unet_chat_info_defaults(PurpleConnection *gc, const char *chat_name);
static void buddy_incoming_im(char *buddyname, char *message, PurpleAccount *acct, time_t mtime);
static time_t strtotime(char *tmpts);
static void send_status(PurpleAccount *acct, int isLogin);
static int prpl_to_unet_status(int status_type);
static char *unet_to_prpl_status(int status_type);
static void buddy_incoming_chat(char *chatname, char *buddyname, char *message, PurpleAccount *acct, time_t mtime);
static void join_chat(char *chatname, PurpleAccount *acct);
static void leave_chat(char *chatname, PurpleAccount *acct);
static void buddy_joined(char *chatname, char *buddyname, PurpleAccount *acct);
static void buddy_left(char *chatname, char *buddyname, PurpleAccount *acct);
static PurpleConvChatBuddyFlags unet_to_prpl_bflags(int buddy_flag);
static void buddy_chat_type(char *chatname, char *buddyname, int btype, PurpleAccount *acct);
static void unet_add_chat_commands();
static PurpleCmdRet chat_add_op(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata);
static PurpleCmdRet chat_remove_op(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata);
static void system_incoming_chat(char *message, PurpleConversation *conv);
static void unet_incoming_system_chat(char *chatname, char *message, PurpleAccount *acct);
static PurpleCmdRet chat_kick_user(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata);
static PurpleCmdRet chat_unban_user(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata);
static PurpleCmdRet chat_ban_user(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata);
static PurpleCmdRet chat_font_formatted(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata);
static void set_chat_format(char *chatname, int format_type, PurpleAccount *acct);
static const char *buddy_normalize(const PurpleAccount *acct, const char *tmpstr);
static int get_next_chat_id();
static PurpleCmdRet chat_mode(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata);
static gboolean disconnect_cb(gpointer data);
static void show_message(const char *msg, const char *submsg, PurpleAccount *acct);
static void show_error(const char *msg, const char *submsg, PurpleAccount *acct);
static PurpleCmdRet chat_history(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata);
static PurpleCmdRet chat_logging(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata);
static void show_approval(const char *user, const char *chat, PurpleAccount *acct);
static void join_approve(ApproveWrapper *tmpwrap);
static void join_reject(ApproveWrapper *tmpwrap);
static void buddy_chat_invite(PurpleConnection *pc, int id, const char *msg, const char *user_name);
static PurpleRoomlist *fetch_roomlist(PurpleConnection *pc);
static void accept_chat_invite(ApproveWrapper *tmpwrap);
static void show_invite(const char *from_user, const char *chat_name, PurpleAccount *acct);
static void unet_incoming_system_im(char *buddyname, char *message, PurpleAccount *acct);
static void myprintf(char *printdata);
static const char *unet_normalize(const char *tmpstr);
static void free_ustr(ustr *tmpstr);
static void free_aw(ApproveWrapper *tmpaw);
static void free_ud(UNetData *ud);
static void show_joinq(const char *msg, const char *submsg, const char *chat_name, PurpleAccount *acct);
static void cancel_roomlist(PurpleRoomlist *tmp_rlist);
static void add_to_roomlist(UNetData *ud, const char *tmpstr);
static void roomlist_complete(UNetData *ud);
static UNetData *get_ud_from_conn(PurpleConnection *pc);
static char *strip_html_string(char *text, char *htmlchars, int addspace);
static PurpleCmdRet chat_font_newlines(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata);
static void set_chat_newlines(char *chatname, int format_type, PurpleAccount *acct);
static PurpleCmdRet chat_invite(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *userdata);
static void open_website(char *site_url);
static char *alt_sha512(PurpleAccount *acct, int hashbytes);
static UNetData *get_ud_from_acct(PurpleAccount *acct);
static int verify_sha512(PurpleAccount *acct, char *testhash, ustr *tmpsalt);
static unsigned char *genran(UNetData *ud, int gensize);
static ustr *pkcs5_strip(ustr *input);
static ustr *pkcs5_pad(ustr *input);
static int pkcs5_len(unsigned char *input, int inputlen);

#define USE_SHA512 0
#define USE_SHA384 1

#define MAX_NORMALIZE	1024
#define CONNECTION_STEPS 5
#define PACKET_HEAD_LEN 2
#define DEFAULT_SERVER 	"login.unfairnet.com"
#define DEFAULT_PORT 	1189
#define NEW_DEFAULT_SERVER	"connect.unfairnet.com"

#define VERSION_MAJOR 	"0"
#define VERSION_MINOR 	"1"
#define VERSION_REV		"6"
#define PROTOCOL_VER	2
#define AUTHOR_USER_ID	1	// This is your unet account id, found by logging into your account at http://unfairnet.com/account/
							// it is important, allowing you to manage versions of your plugin - revoking old buggy versions

// separate our AUTH_NAME for the plugin by platform
#ifdef _WIN32
	#ifndef PLATFORM_LETTER
	#define PLATFORM_LETTER	"W"
	#endif
#endif

#ifdef __APPLE__
	#ifndef PLATFORM_LETTER
	#define PLATFORM_LETTER	"A"
	#endif
#endif

#ifdef __GNUC__
	#ifndef PLATFORM_LETTER // without this second check W turns into L when we cross-compile
	#define PLATFORM_LETTER	"L"
	#endif
#endif

#ifndef PLATFORM_LETTER
#define PLATFORM_LETTER	"U"
#endif

#define UNET_AUTH_NAME	"UPP_" PLATFORM_LETTER // This goes along with the AUTHOR_USER_ID, it is never displayed, only used by
													// the server to determine which plugin, if you have multiple
													// caps and spaces do not matter

#define PLUGIN_ID		"prpl-unfair-unet" // theoretically this should be prpl-unfair-unet but that's not working properly with Empathy and all the other protocols are this format
#define DOWNLOAD_URL	"http://unfairnet.com/download/"

#define PLUGIN_NAME		"UNet"
#define PLUGIN_DESC_S	"Enable connections to UNet"
#define PLUGIN_DESC_L	"Enable connections to UNet - a secure, distributed instant messenger network with centralized authentication"
#define PLUGIN_EMAIL	"Unfair <unfair@unfairnet.com>"
#define PLUGIN_URL		"http://unfairnet.com"

#define UNET_STATUS_OFFLINE	0
#define UNET_STATUS_ONLINE 	1
#define UNET_STATUS_AWAY 	2
#define UNET_STATUS_MOBILE 	3

#define UNET_USER_NORMAL 	0
#define UNET_USER_OP 		1
#define UNET_USER_FOUNDER 	2
#define UNET_USER_ADMIN 	3

int chatcounter = 0;
FILE *debugfile = NULL;
int debugenabled = 0;

struct _UNetData
{
	int current_step;
	PurpleConnection *conn;
	PurpleProxyConnectData *connect_data; // used when connecting, if you need to cancel
	PurpleAccount *acct;
	int fd;
	guint in_fd;
	guint ka_fd;
	guint out_fd;
	guint salt_fd;
	guint dc_fd;
	unsigned char *salt;
	char *sbuff;
	int sbuff_len;
	char *obuff;
	int obuff_len;
	PurpleRoomlist *rlist;
	rsa_context rsapub;
	ustr *tmprsa;
	int symmetric_enabled;
	havege_state rgen;
	aes_context encctx;
	aes_context decctx;
	unsigned char encIV[16];
	unsigned char decIV[16];
};

struct _ustr
{
	int len;
	char *data;
};

struct _ApproveWrapper
{
	PurpleAccount *acct;
	char *chat_name;
	char *user_name;
};

static PurplePluginProtocolInfo proto_info = {
	0, /* options */
	NULL, /* A GList of PurpleAccountUserSplit. */
	NULL, /* A GList of PurpleAccountOption. **REQUIRED, set in init** */
	{"png,jpg,gif", 0, 0, 500, 500, 10000, PURPLE_ICON_SCALE_SEND}, /* The icon spec. */ // ****
	base_icon, /* Returns the base icon name for the given buddy and account. */
	NULL, /* Fills the four char**'s with string identifiers for "emblems" that the UI will interpret and display as relevant. */
	buddy_status, /* Gets a short string representing this buddy's status. **OPENING IMS WILL CRASH IF YOU GIVE A BAD STATUS** */
	buddy_tooltip, /* Allows the prpl to add text to a buddy's tooltip. */
	buddy_status_types, /* Returns a list of PurpleStatusType which exist for this account; this must be implemented, and must add at least the offline and online states. */
	buddy_rclick_menu, /* Returns a list of PurpleMenuAction structs, which represent extra actions to be shown in (for example) the right-click menu for node. */
	unet_chat_info, /* chat_info */
	unet_chat_info_defaults, /* chat_info_defaults */
	user_login, /* This must be implemented. */
	user_logout, /* This must be implemented. */
	buddy_im, /* This PRPL function should return a positive value on success. */
	NULL, /* set_info */
	NULL, /* send_typing */
	buddy_get_info, /* Should arrange for purple_notify_userinfo() to be called with who's user info. */
	buddy_set_status, /* set_status */
	NULL, /* set_idle */
	NULL, /* change_passwd */
	buddy_add, /* add_buddy */
	NULL, /* add_buddies */
	buddy_remove, /* remove_buddy */
	NULL, /* remove_buddies */
	NULL, /* add_permit */
	buddy_block, /* add_deny */
	NULL, /* rem_permit */
	buddy_unblock, /* rem_deny */
	NULL, /* set_permit_deny */
	buddy_join_chat, /* join_chat */
	NULL, /* reject_chat */
	NULL, /* get_chat_name */
	buddy_chat_invite, /* chat_invite */
	buddy_leave_chat, /* chat_leave */
	NULL, /* chat_whisper */
	buddy_chat_im, /* chat_send */
	NULL, /* keepalive - If implemented, this will be called regularly for this prpl's active connections.
			this supposedly uses the socket timeout for frequency, but when that value is set > 0 this doesnt run
			we create our own keepalive with the purple_timeout functions instead. */
	NULL, /* register_user - new user registration */
	NULL, /* get_cb_info */
	NULL, /* get_cb_away */
	NULL, /* alias_buddy - save/store buddy's alias on server list/roster */
	NULL, /* group_buddy - change a buddy's group on a server list/roster */
	NULL, /* rename_group - rename a group on a server list/roster */
	NULL, /* buddy_free */
	NULL, /* convo_closed */
	buddy_normalize, /* normalize - Convert the username who to its canonical form. */
	NULL, /* set_buddy_icon - Set the buddy icon for the given connection to img. */
	NULL, /* remove_group */
	NULL, /* get_cb_real_name - Gets the real name of a participant in a chat. */
	NULL, /* set_chat_topic */
	NULL, /* find_blist_chat */
	fetch_roomlist, /* roomlist_get_list */
	cancel_roomlist, /* roomlist_cancel */
	NULL, /* roomlist_expand_category */
	NULL, /* can_receive_file */
	NULL, /* send_file */
	NULL, /* new_xfer */
	buddy_offline_support, /* offline_message - Checks whether offline messages to buddy are supported. */
	NULL, /* whiteboard_prpl_ops */
	NULL, /* send_raw - For use in plugins that may understand the underlying protocol. */
	NULL, /* roomlist_room_serialize */
	NULL, /* unregister_user - Remove the user from the server. */
	NULL, /* send_attention */
	NULL, /* get_attention_types */
	sizeof(PurplePluginProtocolInfo), /* The size of the PurplePluginProtocolInfo. */
	NULL /* This allows protocols to specify additional strings to be used for various purposes.  */
};

static PurplePluginInfo info = {
    PURPLE_PLUGIN_MAGIC,    /* Plugin magic, this must always be
                               PURPLE_PLUGIN_MAGIC.*/

    PURPLE_MAJOR_VERSION,   /* This is also defined in libpurple.  It helps
                               libpurple's plugin system determine which version
                               of libpurple this plugin was compiled for, and
                               whether loading it will cause problems. */

    PURPLE_MINOR_VERSION,   /* See previous */

    PURPLE_PLUGIN_PROTOCOL, /* PurplePluginType: There are 4 different values for
                               this field.  The first is PURPLE_PLUGIN_UNKNOWN,
                               which should not be used.  The second is
                               PURPLE_PLUGIN_STANDARD; this is the value most
                               plugins will use. Next, we have PURPLE_PLUGIN_LOADER;
                               this is the type you want to use if your plugin will
                               make it possible to load non-native plugins.  For
                               example, the Perl and Tcl loader plugins are of this
                               type.  Last, we have PURPLE_PLUGIN_PROTOCOL.  If your
                               plugin is going to allow the user to connect to
                               another network, this is the type you'd want to use. */

    NULL,                   /* This field is the UI requirement.  If you're writing
                               a core plugin, this must be NULL and the plugin must
                               not contain any UI code.  If you're writing a Pidgin
                               plugin, you need to use PIDGIN_PLUGIN_TYPE.  If you
                               are writing a Finch plugin, you would use
                               FINCH_PLUGIN_TYPE.*/

    0,                      /* This field is for plugin flags.  Currently, the only
                               flag available to plugins is invisible
                               (PURPLE_PLUGIN_FLAG_INVISIBLE). It causes the plugin
                               NOT to appear in the list of plugins. */

    NULL,                   /* This is a GList of plugin dependencies.  In other words,
                               it's a GList of plugin id's that your plugin depends on.
                               Set this value to NULL no matter what.  If your plugin
                               has dependencies, set them at run-time in the
                               plugin_init function. */

    PURPLE_PRIORITY_DEFAULT,/* This is the priority libpurple will give your plugin.
                               There are three possible values for this field,
                               PURPLE_PRIORITY_DEFAULT, PURPLE_PRIORITY_HIGHEST, and
                               PURPLE_PRIORITY_LOWEST. */

    PLUGIN_ID,     /* This is your plugin's id.  There is a whole page dedicated
                               to this in the Related Pages section of the API docs. */

    PLUGIN_NAME,         /* This is your plugin's name.  This is what will be
                               displayed for your plugin in the UI. */

    VERSION_MAJOR "." VERSION_MINOR "." VERSION_REV,  /* This is the version of your plugin. */

    PLUGIN_DESC_S,   /* This is the summary of your plugin.  It should be a short
                               blurb.  The UI determines where, if at all, to display
                               this. */

    PLUGIN_DESC_L,   /* This is the description of your plugin. It can be as long
                               and as descriptive as you like.  And like the summary,
                               it's up to the UI where, if at all, to display this (and
                               how much to display). */

    PLUGIN_EMAIL, /* This is where you can put your name and e-mail
                                         address. */

    PLUGIN_URL,/* This is the website for the plugin.  This tells users
                               where to find new versions, report bugs, etc. */

    plugin_load,            /* This is a pointer to a function for libpurple to call when
                               it is loading the plugin.  It should be of the type:

                                  gboolean plugin_load(PurplePlugin *plugin)

                               Returning FALSE will stop the loading of the plugin.
                               Anything else would evaluate as TRUE and the plugin will
                               continue to load. */

    plugin_unload,                   /* Same as above except it is called when libpurple tries to
                               unload your plugin.  It should be of the type:

                                  gboolean plugin_unload(PurplePlugin *plugin)

                               Returning TRUE will tell libpurple to continue unloading
                               while FALSE will stop the unloading of your plugin. */

    destroy_plugin,                   /* Similar to the two above members, except this is called
                               when libpurple tries to destory the plugin.  This is
                               generally only called when for some reason or another the
                               plugin fails to probe correctly.  It should be of the type:

                                  void plugin_destroy(PurplePlugin *plugin) */

    NULL,                   /* This is a pointer to a UI-specific struct.  For a Pidgin
                               plugin it will be a pointer to a PidginPluginUiInfo
                               struct, for example. */

    &proto_info,                   /* This is a pointer to either a PurplePluginLoaderInfo
                               struct or a PurplePluginProtocolInfo struct, and is
                               beyond the scope of this document.
                               NOTE: this is REQUIRED for a protocol plugin */

    NULL,                   /* This is a pointer to a PurplePluginUiInfo struct.  It is
                               a core/ui split way for core plugins to have a UI
                               configuration frame.  You can find an example of this
                               code in libpurple/plugins/pluginpref_example.c */

    NULL,                   /* This is a function pointer where you can define "plugin
                               actions".  The UI controls how they're displayed.  It
                               should be of the type:

                                  GList *function_name(PurplePlugin *plugin,
                                                       gpointer context)

                               It must return a GList of PurplePluginActions. */

    NULL,                   /* This is a pointer reserved for future use.  We set it to
                               NULL to indicate we don't need it. */

    NULL,                   /* This is a pointer reserved for future use.  We set it to
                               NULL to indicate we don't need it. */

    NULL,                   /* This is a pointer reserved for future use.  We set it to
                               NULL to indicate we don't need it. */

    NULL                    /* This is a pointer reserved for future use.  We set it to
                               NULL to indicate we don't need it. */
};

PURPLE_INIT_PLUGIN(unet, init_plugin, info);
